// Author: A.Voss@FH-Aachen.de
//
// smart ptrs
// tags: shared_ptr, weak_ptr

#include <iostream>
#include <memory>
#include <cstdlib>
#include <vector>
using namespace std;

#include "../../include/tools.h"

// http://ootips.org/yonat/4dev/smart-pointers.html

/* from cppreference:
    std::shared_ptr is a smart pointer that retains shared ownership of an object 
    through a pointer. Several shared_ptr objects may own the same object. The 
    object is destroyed and its memory deallocated when either of the following 
    happens:
        the last remaining shared_ptr owning the object is destroyed.
        the last remaining shared_ptr owning the object is assigned another 
        pointer via operator= or reset().
        The object is destroyed using delete-expression or a custom deleter that 
        is supplied to shared_ptr during construction.
        A shared_ptr can share ownership of an object while storing a pointer to 
        another object. This feature can be used to point to member objects while 
        owning the object they belong to.
        A shared_ptr may also own no objects, in which case it is called empty.
    
    std::weak_ptr is a smart pointer that holds a non-owning ("weak") reference 
    to an object that is managed by std::shared_ptr. It must be converted to 
    std::shared_ptr in order to access the referenced object.
    std::weak_ptr models temporary ownership: when an object needs to be accessed 
    only if it exists, and it may be deleted at any time by someone else, 
    std::weak_ptr is used to track the object, and it is converted to 
    std::shared_ptr to assume temporary ownership. If the original std::weak_ptr 
    is destroyed at this time, the object's lifetime is extended until the 
    temporary std::shared_ptr is destroyed as well.
    In addition, std::weak_ptr is used to break circular references of std::shared_ptr.
*/

class C {
public:
    int n;
    
    C(int N=0) : n(N) { 
        tools_log(); cout << "ctor " << n << endl; 
    }
    ~C() { 
        tools_log(); cout << "dtor " << n << endl; 
    }
    
    friend ostream & operator<<(ostream & os, const C & c) { 
        os << c.n; return os; 
    }
};

// some factory function
C * createC(int n) { return new C(n); }

class D : public C {
public:
    typedef shared_ptr<D> Ptr;
    Ptr p;

    D(int N=0) : C(N) { }
    
    friend ostream & operator<<(ostream & os, const D & c) { os << c.n; return os; }
};

int main()
{

    {
    tools_log(); cout << "shared ptr usage..." << endl;

    shared_ptr<C> p1a( createC(1) );   // first ref 
    shared_ptr<C> p1b = p1a;           // second ref 
       
    tools_log(); cout << "*p1a:" << (*p1a) << " p1a:" << p1a.get() << " use_cnt: " << p1a.use_count() << endl;
    tools_log(); cout << "*p1b:" << (*p1b) << " p1b:" << p1b.get() << " use_cnt: " << p1b.use_count() << endl;

    p1a.reset();                       // nothing happens
    tools_log(); cout << "*p1a:" << "-"    << " p1a:" << p1a.get() << " use_cnt: " << p1a.use_count() << endl;
    tools_log(); cout << "*p1b:" << (*p1b) << " p1b:" << p1b.get() << " use_cnt: " << p1b.use_count() << endl;

    p1b.reset();                       // mem is gone
    tools_log(); cout << "*p1a:" << "-"    << " p1a:" << p1a.get() << " use_cnt: " << p1a.use_count() << endl;
    tools_log(); cout << "*p1b:" << "-"    << " p1b:" << p1b.get() << " use_cnt: " << p1b.use_count() << endl;
    }
    
    {
    cout << endl;
    tools_log(); cout << "shared and weak ptr usage..." << endl;

    shared_ptr<C> p2a( createC(2) );
    weak_ptr<C>   w2 = p2a;
    tools_log(); cout << "*p2a:" << (*p2a) << " p2a:" << p2a.get() << " use_cnt: " << p2a.use_count() << endl;

    // data access via *w2 not possible, a second shared_ptr is needed

    shared_ptr<C> p2b = w2.lock();
    tools_log(); cout << "*p2a:" << (*p2a) << " p2a:" << p2a.get() << " use_cnt: " << p2a.use_count() << endl;
    tools_log(); cout << "*p2b:" << (*p2b) << " p2b:" << p2b.get() << " use_cnt: " << p2b.use_count() << endl;
       
    // reset ptr, mem is gone even a weak ptr points to it
    p2a.reset();  
    p2b.reset();  
    shared_ptr<C> p2c = w2.lock();
    tools_log(); cout << "*p2c:" << "-"    << " p2c:" << p2c.get() << " use_cnt: " << p2c.use_count() << endl << endl;
    
    shared_ptr<C> p3 (new C(3));
    shared_ptr<C> p4 = p3;
    tools_log(); cout << "p3==p4? " << (p3==p4) << endl;
    }
    
    {
    cout << endl;
    tools_log(); cout << "show circular..." << endl;

    D::Ptr p1 (new D(1));
    D::Ptr p2 (new D(2));

    tools_log(); cout << "*p1:" << (*p1) << " use_cnt: " << p1.use_count() << endl;
    tools_log(); cout << "*p2:" << (*p2) << " use_cnt: " << p2.use_count() << endl;
    
    p1->p = p2;
    p2->p = p1;

    tools_log(); cout << "*p1:" << (*p1) << " use_cnt: " << p1.use_count() << endl;
    tools_log(); cout << "*p2:" << (*p2) << " use_cnt: " << p2.use_count() << endl;
    }
    
    return EXIT_SUCCESS;
}
